# Uses code from https://github.com/python/cpython/pull/111515
# type: ignore
import collections.abc
import functools
import operator
import sys
import types
from collections import defaultdict
from types import GenericAlias
from typing import (
    Any,
    ForwardRef,
    Generic,
    TypeVar,
    Union,
    _AnnotatedAlias,
    _eval_type,
    _GenericAlias,
    _type_check,
    get_args,
    get_origin,
)

from typing_extensions import NotRequired, ParamSpec, Required, TypeVarTuple, is_typeddict

_allowed_types = (
    types.FunctionType,
    types.BuiltinFunctionType,
    types.MethodType,
    types.ModuleType,
    types.WrapperDescriptorType,
    types.MethodWrapperType,
    types.MethodDescriptorType,
)


def get_type_hints(
    obj: Any,
    globalns: Union[dict[str, Any], None] = None,
    localns: Union[dict[str, Any], None] = None,
    include_extras: bool = False,
) -> dict[str, Any]:
    """Return type hints for an object.

    This is often the same as obj.__annotations__, but it handles
    forward references encoded as string literals, recursively replaces all
    'Annotated[T, ...]' with 'T' (unless 'include_extras=True'), and resolves
    type variables to their values as needed.

    The argument may be a module, class, generic alias, method, or function.
    The annotations are returned as a dictionary. For classes and generic aliases,
    annotations also include inherited members.

    TypeError is raised if the argument is not of a type that can contain
    annotations, and an empty dictionary is returned if no annotations are
    present.

    BEWARE -- the behavior of globalns and localns is counterintuitive
    (unless you are familiar with how eval() and exec() work).  The
    search order is locals first, then globals.

    - If no dict arguments are passed, an attempt is made to use the
      globals from obj (or the respective module's globals for classes),
      and these are also used as the locals.  If the object does not appear
      to have globals, an empty dictionary is used.  For classes, the search
      order is globals first then locals.

    - If one dict argument is passed, it is used for both globals and
      locals.

    - If two dict arguments are passed, they specify globals and
      locals, respectively.
    """
    if getattr(obj, '__no_type_check__', None):
        return {}

    # for Generic Aliases we need to inspect the origin
    # then apply its args later
    ga_args = None
    if isinstance(obj, (_GenericAlias, GenericAlias)):
        ga_args = get_args(obj)
        obj = get_origin(obj)

    # Classes require a special treatment.
    if isinstance(obj, type):
        # track typevars of each base
        param_tracking = defaultdict(list)
        # track type hints of each base
        hint_tracking = {}
        hints = {}
        # typeddicts cannot redefine pre-existing keys
        can_override = not is_typeddict(obj)
        for base in _get_all_bases(obj):
            # keep track of typevars and the values they are being
            # replaced with
            for cls, args in _track_parameter_changes(base):
                param_tracking[cls].append(args)

                if cls is not base:
                    # if we previously scanned it and it found no type hints
                    # then skip processing it
                    if not hint_tracking[cls]:
                        continue

                    to_sub = _substitute_type_hints(param_tracking[cls], hint_tracking[cls])
                    hints.update(to_sub)

            if globalns is None:
                base_globals = getattr(sys.modules.get(base.__module__, None), '__dict__', {})
            else:
                base_globals = globalns
            ann = base.__dict__.get('__annotations__', {})
            if isinstance(ann, types.GetSetDescriptorType):
                ann = {}
            base_locals = dict(vars(base)) if localns is None else localns
            if localns is None and globalns is None:
                # This is surprising, but required.  Before Python 3.10,
                # get_type_hints only evaluated the globalns of
                # a class.  To maintain backwards compatibility, we reverse
                # the globalns and localns order so that eval() looks into
                # *base_globals* first rather than *base_locals*.
                # This only affects ForwardRefs.
                base_globals, base_locals = base_locals, base_globals

            hint_tracking[base] = {}
            for name, value in ann.items():
                # skip pre-existing keys for typeddict
                if not can_override and name in hints:
                    continue

                if value is None:
                    value = type(None)
                if isinstance(value, str):
                    value = ForwardRef(value, is_argument=False, is_class=True)
                value = _eval_type(value, base_globals, base_locals)
                hint_tracking[base][name] = value
                hints[name] = value

        # sub the original args back in
        if ga_args is not None:
            param_tracking[obj].append(ga_args)
            to_sub = _substitute_type_hints(param_tracking[obj], hints)
            hints.update(to_sub)

        return hints if include_extras else {k: _strip_annotations(t) for k, t in hints.items()}

    if globalns is None:
        if isinstance(obj, types.ModuleType):
            globalns = obj.__dict__
        else:
            nsobj = obj
            # Find globalns for the unwrapped object.
            while hasattr(nsobj, '__wrapped__'):
                nsobj = nsobj.__wrapped__
            globalns = getattr(nsobj, '__globals__', {})
        if localns is None:
            localns = globalns
    elif localns is None:
        localns = globalns
    hints = getattr(obj, '__annotations__', None)
    if hints is None:
        # Return empty annotations for something that _could_ have them.
        if isinstance(obj, _allowed_types):
            return {}
        else:  # noqa: RET505
            raise TypeError(f'{obj!r} is not a module, class, method, ' 'or function.')
    hints = dict(hints)
    for name, value in hints.items():
        if value is None:
            value = type(None)
        if isinstance(value, str):
            # class-level forward refs were handled above, this must be either
            # a module-level annotation or a function argument annotation
            value = ForwardRef(
                value,
                is_argument=not isinstance(obj, types.ModuleType),
                is_class=False,
            )
        hints[name] = _eval_type(value, globalns, localns)
    return hints if include_extras else {k: _strip_annotations(t) for k, t in hints.items()}


def _substitute_type_hints(substitutions: 'list[tuple[Any, ...]]', hints: 'dict[str, Any]'):
    # nothing to substitute
    if len(substitutions) < 2:
        return {}

    previous_params = substitutions[-2]
    new_params = substitutions[-1]
    new_substitutions = _repack_args(previous_params, new_params)

    # get a mapping of typevar to value
    mapping = {}
    for i in range(len(previous_params)):
        current = previous_params[i]
        # if the previous parameter is *Ts
        # make it {Ts: new Ts or value}
        if _is_unpacked_typevartuple(current):
            (previous_typevartuple,) = get_args(current)
            new_value = new_substitutions[i]

            if _is_unpacked_typevartuple(new_value):
                (new_value,) = get_args(new_value)

            mapping[previous_typevartuple] = new_value
        else:
            mapping[current] = new_substitutions[i]

    hints_to_replace = {}

    for name, value in hints.items():
        origin = get_origin(value)
        # if the typevar is nested, we must substitute the typevar all the way down.
        if origin is not None:
            new_args = tuple(_make_substitution(origin, get_args(value), mapping))
            sub = _copy_with(value, new_args)
        elif _is_typevar_like(value):
            # https://github.com/python/cpython/pull/111515#issuecomment-2018336687
            # sub = mapping[value]  # old
            sub = mapping.get(value, value)
        else:
            continue

        hints_to_replace[name] = sub

    return hints_to_replace


def _copy_with(t, new_args):
    if new_args == t.__args__:
        return t
    if isinstance(t, GenericAlias):
        return GenericAlias(t.__origin__, new_args)
    if isinstance(t, _AnnotatedAlias):  # https://github.com/python/cpython/pull/111515#issuecomment-2018132920
        return t.copy_with(new_args[:1])
    if hasattr(types, 'UnionType') and isinstance(t, types.UnionType):
        return functools.reduce(operator.or_, new_args)
    else:  # noqa: RET505
        return t.copy_with(new_args)


def _make_substitution(origin, args, new_arg_by_param):
    """Create a list of new type arguments."""
    new_args = []
    for old_arg in args:
        if isinstance(old_arg, type):
            new_args.append(old_arg)
            continue

        substfunc = getattr(old_arg, '__typing_subst__', None)
        if substfunc:
            new_arg = substfunc(new_arg_by_param[old_arg])
        elif isinstance(old_arg, TypeVar) and sys.version_info[:2] <= (3, 10):
            new_arg = _typevar_subst(new_arg_by_param[old_arg])
        else:
            subparams = getattr(old_arg, '__parameters__', ())
            if not subparams:
                new_arg = old_arg
            else:
                subargs = []
                for x in subparams:
                    if isinstance(x, TypeVarTuple):
                        subargs.extend(new_arg_by_param[x])
                    else:
                        subargs.append(new_arg_by_param[x])
                new_arg = old_arg[tuple(subargs)]

        if origin == collections.abc.Callable and isinstance(new_arg, tuple):
            # Consider the following `Callable`.
            #   C = Callable[[int], str]
            # Here, `C.__args__` should be (int, str) - NOT ([int], str).
            # That means that if we had something like...
            #   P = ParamSpec('P')
            #   T = TypeVar('T')
            #   C = Callable[P, T]
            #   D = C[[int, str], float]
            # ...we need to be careful; `new_args` should end up as
            # `(int, str, float)` rather than `([int, str], float)`.
            new_args.extend(new_arg)
        elif _is_unpacked_typevartuple(old_arg):
            # Consider the following `_GenericAlias`, `B`:
            #   class A(Generic[*Ts]): ...
            #   B = A[T, *Ts]
            # If we then do:
            #   B[float, int, str]
            # The `new_arg` corresponding to `T` will be `float`, and the
            # `new_arg` corresponding to `*Ts` will be `(int, str)`. We
            # should join all these types together in a flat list
            # `(float, int, str)` - so again, we should `extend`.
            new_args.extend(new_arg)
        elif isinstance(old_arg, tuple):
            # Corner case:
            #    P = ParamSpec('P')
            #    T = TypeVar('T')
            #    class Base(Generic[P]): ...
            # Can be substituted like this:
            #    X = Base[[int, T]]
            # In this case, `old_arg` will be a tuple:
            new_args.append(
                tuple(_make_substitution(origin, old_arg, new_arg_by_param)),
            )
        else:
            new_args.append(new_arg)
    return new_args


def _repack_args(reference, params):
    """transforms params to match the requested arguments
    of reference.

    >>> _repack_args((T, U, *Ts), (int, str, float, bool))
    (int, str, (float, bool))
    >>> _repack_args((T, U, *Ts), (int, str, *Ts))
    (int, str, *Ts)
    """

    # find bounds of potential typevar tuple
    tuple_start, tuple_end = 0, None
    found_typvartuple = False
    for i, tv in enumerate(reference):
        if _is_unpacked_typevartuple(tv) or isinstance(tv, TypeVarTuple):
            tuple_start = i
            found_typvartuple = True
        elif found_typvartuple:
            tuple_end = i
            break

    if found_typvartuple:
        # if typevartuple doesnt consume rest
        if tuple_end is not None:
            tuple_end = tuple_end - len(reference)

        # tuple of typevars
        type_var_tuple_params = params[tuple_start:tuple_end]

        # if params might contain typevartuple args
        if len(params) > tuple_start:  # noqa: SIM102
            # if params[tuple_start] is *Ts keep it as-is
            if _is_unpacked_typevartuple(params[tuple_start]):
                type_var_tuple_params = params[tuple_start]

        # if starts with type var tuple
        if tuple_start == 0:
            # if there are type vars after it
            if tuple_end is not None:
                return (type_var_tuple_params, *params[tuple_end:])
            return (type_var_tuple_params,)
        # if it's at the end
        elif tuple_end is None:  # noqa: RET505
            return (*params[:tuple_start], type_var_tuple_params)
        # if it's in the middle
        return (*params[:tuple_start], type_var_tuple_params, *params[tuple_end:])
    return params


def _is_unpacked_typevartuple(x: Any) -> bool:
    return (not isinstance(x, type)) and getattr(x, '__typing_is_unpacked_typevartuple__', False)


def _is_typevar_like(x: Any) -> bool:
    return isinstance(x, (TypeVar, ParamSpec)) or _is_unpacked_typevartuple(x)


def _get_all_bases(cls):
    """Correctly obtains the bases for classes
    and typeddicts alike.
    """
    mro = reversed(cls.__mro__)

    if not is_typeddict(cls):
        yield from mro
        return

    traversing = list(mro)
    visited = []

    while traversing:
        base = traversing.pop(0)

        orig_bases = getattr(base, '__orig_bases__', ())
        mro_changed = False

        for orig_base in orig_bases:
            origin = get_origin(orig_base)

            if origin is not Generic:
                new_base = orig_base if origin is None else origin
                if new_base not in visited:
                    traversing.insert(0, new_base)
                    traversing.insert(1, base)
                    mro_changed = True

        if mro_changed:
            continue

        yield base

        visited.append(base)


def _track_parameter_changes(base):
    """Track how a parameters values are substituted
    or changed while traversing its bases.
    """
    orig_bases = getattr(base, '__orig_bases__', ())
    generic_encountered = False

    for orig_base in orig_bases:
        origin = get_origin(orig_base)
        if origin is None:
            continue

        args = get_args(orig_base)

        if origin is Generic:
            generic_encountered = True
            yield base, args
        else:
            yield origin, args

    # this occurs if obj is
    # class Bar(Foo[str, U]): ...
    # in this case, we need to imagine there's a generic base
    # with the required typevars here.
    if orig_bases and not generic_encountered:
        # we need to collect all the typevars from all bases
        # in the case that they have multiple generic bases
        # class Baz(Foo[str, U], Bar[U, T]): ...
        type_vars_for_generic = _collect_parameters(orig_bases)

        # this may be empty if obj is
        # class Bar(Foo[str]): ...
        # we can skip adding typevars here.
        if type_vars_for_generic:
            yield base, type_vars_for_generic


def _collect_parameters(args):
    """Collect all type variables and parameter specifications in args
    in order of first appearance (lexicographic order).

    For example::

        assert _collect_parameters((T, Callable[P, T])) == (T, P)
    """
    parameters = []
    for t in args:
        if isinstance(t, type):
            # We don't want __parameters__ descriptor of a bare Python class.
            pass
        elif isinstance(t, tuple):
            # `t` might be a tuple, when `ParamSpec` is substituted with
            # `[T, int]`, or `[int, *Ts]`, etc.
            for x in t:
                for collected in _collect_parameters([x]):
                    if collected not in parameters:
                        parameters.append(collected)
        elif hasattr(t, '__typing_subst__'):
            if t not in parameters:
                parameters.append(t)
        else:
            for x in getattr(t, '__parameters__', ()):
                if x not in parameters:
                    parameters.append(x)
    return tuple(parameters)


def _strip_annotations(t):
    """Strip the annotations from a given type."""
    if isinstance(t, _AnnotatedAlias):
        return _strip_annotations(t.__origin__)
    if hasattr(t, '__origin__') and t.__origin__ in (Required, NotRequired):
        return _strip_annotations(t.__args__[0])
    if isinstance(t, _GenericAlias):
        stripped_args = tuple(_strip_annotations(a) for a in t.__args__)
        if stripped_args == t.__args__:
            return t
        return t.copy_with(stripped_args)
    if isinstance(t, GenericAlias):
        stripped_args = tuple(_strip_annotations(a) for a in t.__args__)
        if stripped_args == t.__args__:
            return t
        return GenericAlias(t.__origin__, stripped_args)
    if hasattr(types, 'UnionType') and isinstance(t, types.UnionType):
        stripped_args = tuple(_strip_annotations(a) for a in t.__args__)
        if stripped_args == t.__args__:
            return t
        return functools.reduce(operator.or_, stripped_args)

    return t


# python 3.9 / 3.10 compatibility


def _typevar_subst(arg):
    msg = 'Parameters to generic types must be types.'
    arg = _type_check(arg, msg, is_argument=True)
    return arg



import sys
from typing import Any


def get_dataclass_args() -> dict[str, Any]:
    # python 3.9 does not support dataclasses(slot=True)
    if sys.version_info < (3, 10):
        return {}
    return {"slots": True}



import pytest

from .libs import marshmallow, mashumaro, pydantic, serpyco, serpyco_rs

serializers = {
    'serpyco_rs': serpyco_rs,
    'serpyco': serpyco,
    'pydantic': pydantic,
    'marshmallow': marshmallow,
    'mashumaro': mashumaro,
}


@pytest.mark.parametrize('lib', serializers.keys())
def test_dump(bench_or_check_refcount, lib):
    serializer = serializers[lib]
    serializer.dump(serializer.test_object)  # warmup

    bench_or_check_refcount.group = 'dump'
    bench_or_check_refcount.extra_info['lib'] = lib
    bench_or_check_refcount.extra_info['correct'] = (
        serializer.load(serializer.dump(serializer.test_object)) == serializer.test_object
    )
    if lib in {'serpyco', 'pydantic'}:
        bench_or_check_refcount.skip_refcount = True
    bench_or_check_refcount(serializer.dump, serializer.test_object)


@pytest.mark.parametrize('lib', serializers.keys())
def test_load_validate(bench_or_check_refcount, lib):
    serializer = serializers[lib]
    test_dict = serializer.dump(serializer.test_object)
    serializer.load(test_dict)  # warmup

    bench_or_check_refcount.group = 'load with validate'
    bench_or_check_refcount.extra_info['lib'] = lib
    bench_or_check_refcount.extra_info['correct'] = (
        serializer.load(serializer.dump(serializer.test_object)) == serializer.test_object
    )
    if lib in {'serpyco', 'pydantic'}:
        bench_or_check_refcount.skip_refcount = True
    bench_or_check_refcount(serializer.load, test_dict)



from typing import TypeVar, Generic, TypeVarTuple, Tuple, Unpack

from serpyco_rs._type_utils import get_type_hints

T = TypeVar('T')
U = TypeVar('U')
KT = TypeVar('KT')
VT = TypeVar('VT')


def test_get_type_hints_variadic_class():
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')
    Ts = TypeVarTuple('Ts')

    class A(Generic[T1, T2, *Ts]):
        t1: T1
        t2: T2
        ts: Tuple[*Ts]

    assert get_type_hints(A[int, str]) == {'t1': int, 't2': str, 'ts': Tuple[()]}
    assert get_type_hints(A[int, str, float]) == {'t1': int, 't2': str, 'ts': Tuple[float]}
    assert get_type_hints(A[int, str, float, bool]) == {'t1': int, 't2': str, 'ts': Tuple[float, bool]}

    class B(Generic[T1, T2, Unpack[Ts]]):
        t1: T1
        t2: T2
        ts: Tuple[Unpack[Ts]]

    assert get_type_hints(B[int, str]) == {'t1': int, 't2': str, 'ts': Tuple[()]}
    assert get_type_hints(B[int, str, float]) == {'t1': int, 't2': str, 'ts': Tuple[float]}
    assert get_type_hints(B[int, str, float, bool]) == {'t1': int, 't2': str, 'ts': Tuple[float, bool]}

    class C(Generic[T1, *Ts, T2]):
        t1: T1
        ts: Tuple[*Ts]
        t2: T2

    assert get_type_hints(C[int, str]) == {'t1': int, 'ts': Tuple[()], 't2': str}
    assert get_type_hints(C[int, str, float]) == {'t1': int, 'ts': Tuple[str], 't2': float}
    assert get_type_hints(C[int, str, float, bool]) == {'t1': int, 'ts': Tuple[str, float], 't2': bool}

    class E(Generic[*Ts, T1, T2]):
        ts: Tuple[*Ts]
        t1: T1
        t2: T2

    assert get_type_hints(E[int, str]) == {'ts': Tuple[()], 't1': int, 't2': str}
    assert get_type_hints(E[int, str, float]) == {'ts': Tuple[int], 't1': str, 't2': float}
    assert get_type_hints(E[int, str, float, bool]) == {'ts': Tuple[int, str], 't1': float, 't2': bool}


def test_get_type_hints_variadic_type_parameter():
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')
    Ts = TypeVarTuple('Ts')

    class A(Generic[T1, T2, *Ts]):
        t1: T1
        t2: T2
        ts: Tuple[*Ts]

    class B(A[int, str, *Ts]):
        pass

    assert get_type_hints(B) == {'t1': int, 't2': str, 'ts': Tuple[*Ts]}
    assert get_type_hints(B[()]) == {'t1': int, 't2': str, 'ts': Tuple[()]}
    assert get_type_hints(B[float]) == {'t1': int, 't2': str, 'ts': Tuple[float]}


def test_get_type_hints_variadic_with_typevars():
    T1 = TypeVar('T1')
    T2 = TypeVar('T2')
    Ts = TypeVarTuple('Ts')

    class A(Generic[T1, T2, *Ts]):
        t1: T1
        t2: T2
        ts: Tuple[*Ts]

    class D(A[int, str, T1, T2]):
        pass

    assert get_type_hints(D[int, str]) == {'t1': int, 't2': str, 'ts': Tuple[int, str]}



import gc
import sys

import pytest


def pytest_addoption(parser):
    parser.addoption(
        '--debug-refs',
        action="store_true",
        default=False,
        help='Check refcount of gc tracked objects before and after the test.',
    )
    parser.addoption(
        '--debug-refs-gc', action="store_true", default=False, help='Collect garbage before and after the test.'
    )


@pytest.fixture
def debug_refs(request):
    return request.config.getoption('--debug-refs')


@pytest.fixture
def debug_refs_gc(request):
    return request.config.getoption('--debug-refs-gc')


@pytest.fixture
def bench_or_check_refcount(benchmark, debug_refs, debug_refs_gc):
    if not debug_refs:
        return benchmark
    benchmark._mode = 'skip benchmark'

    if not hasattr(sys, 'gettotalrefcount'):
        raise RuntimeError('For use --debug-refs option, you need to run tests with python debug build.')

    def inner(fn, *args, tolerance=5, **kwargs):
        if debug_refs_gc:
            gc.collect()
        before = sys.gettotalrefcount()
        fn(*args, **kwargs)
        if debug_refs_gc:
            gc.collect()
        after = sys.gettotalrefcount()
        diff = after - before

        if abs(diff) > tolerance:
            message = f'[refcount changed] before: {before} after: {after} diff: {diff}'
            if inner.skip_refcount:
                pytest.skip(message)
            else:
                pytest.fail(message)

    inner.skip_refcount = False

    # for pytest-benchmark compatibility
    inner.extra_info = {}

    return inner



import collections
import io
import json
import os

from tabulate import tabulate


LIBRARIES = (
    'serpyco_rs',
    'serpyco',
    'mashumaro',
    'pydantic',
    'marshmallow',
)


def aggregate():
    benchmarks_dir = os.path.join('.benchmarks', os.listdir('.benchmarks')[0])
    res = collections.defaultdict(dict)
    for filename in os.listdir(benchmarks_dir):
        with open(os.path.join(benchmarks_dir, filename)) as fileh:
            data = json.loads(fileh.read())

        for each in data['benchmarks']:
            res[each['group']][each['extra_info']['lib']] = {
                'data': [val * 1000 for val in each['stats']['data']],
                'median': each['stats']['median'] * 1000,
                'ops': each['stats']['ops'],
                'correct': each['extra_info']['correct'],
            }
    return res


def tab(obj):
    buf = io.StringIO()
    headers = (
        'Library',
        'Median latency (milliseconds)',
        'Operations per second',
        'Relative (latency)',
    )
    for group, val in sorted(obj.items(), reverse=True):
        buf.write('\n' + '#### ' + group + '\n\n')
        table = []
        for lib in LIBRARIES:
            correct = val[lib]['correct']
            table.append(
                [
                    lib,
                    val[lib]['median'] if correct else None,
                    '%.1f' % val[lib]['ops'] if correct else None,
                    0,
                ]
            )
        baseline = table[0][1]
        for each in table:
            each[3] = '%.2f' % (each[1] / baseline) if isinstance(each[1], float) else None
            each[1] = '%.2f' % each[1] if isinstance(each[1], float) else None
        buf.write(tabulate(table, headers, tablefmt='github') + '\n')

    print(buf.getvalue())


tab(aggregate())



import uuid
from dataclasses import dataclass
from datetime import date, datetime, time
from decimal import Decimal
from enum import Enum
from typing import Annotated, Any, Literal, TypedDict, Union

import pytest
from serpyco_rs import SchemaValidationError, Serializer
from serpyco_rs._impl import ErrorItem
from serpyco_rs.metadata import CustomEncoder, Discriminator, Max, MaxLength, Min, MinLength


def _check_errors(s: Serializer, value: Any, expected_errors: list[ErrorItem]):
    with pytest.raises(SchemaValidationError) as native_schema_error:
        s.load(value)

    assert native_schema_error.value.errors == expected_errors


@pytest.mark.parametrize(
    ['value', 'err'],
    [
        (1, '1 is less than the minimum of 10'),
        (101, '101 is greater than the maximum of 100'),
    ],
)
def test_integer_validation(value, err):
    s = Serializer(Annotated[int, Min(10), Max(100)])
    _check_errors(s, value, [ErrorItem(message=err, instance_path='')])


def test_integer_validation__invalid_type():
    s = Serializer(int)
    _check_errors(s, '1', [ErrorItem(message='"1" is not of type "integer"', instance_path='')])


def test_integer__custom_encoder():
    s = Serializer(Annotated[int, CustomEncoder(serialize=lambda x: x + 1, deserialize=lambda x: x - 1)])
    assert s.dump(10) == 11
    assert s.load(10) == 9


@pytest.mark.parametrize(
    ['value', 'err'],
    [
        ('hello', '"hello" is shorter than 6 characters'),
        ('hello world', '"hello world" is longer than 8 characters'),
    ],
)
def test_string_validation(value, err):
    s = Serializer(Annotated[str, MinLength(6), MaxLength(8)])
    _check_errors(s, value, [ErrorItem(message=err, instance_path='')])


def test_string_validation__invalid_type():
    s = Serializer(str)
    _check_errors(s, 1, [ErrorItem(message='1 is not of type "string"', instance_path='')])


@pytest.mark.parametrize(
    ['value', 'err'],
    [
        (1.1, '1.1 is less than the minimum of 10'),
        (101.2, '101.2 is greater than the maximum of 100'),
        (1, '1 is less than the minimum of 10'),
        (101, '101 is greater than the maximum of 100'),
    ],
)
def test_float_validation(value, err):
    s = Serializer(Annotated[float, Min(10.0), Max(100.0)])

    with pytest.raises(SchemaValidationError) as e:
        s.load(value)

    assert e.value.errors == [ErrorItem(message=err, instance_path='')]


def test_float_validation__invalid_type():
    s = Serializer(float)
    with pytest.raises(SchemaValidationError) as e:
        s.load('1.1')

    assert e.value.errors == [ErrorItem(message='"1.1" is not of type "number"', instance_path='')]


@pytest.mark.parametrize(
    ['value', 'err'],
    [
        ('1.1', '1.1 is less than the minimum of 10'),
        ('101.2', '101.2 is greater than the maximum of 100'),
    ],
)
def test_decimal_validation(value, err):
    s = Serializer(Annotated[Decimal, Min(10), Max(100)])

    with pytest.raises(SchemaValidationError) as e:
        s.load(value)

    assert e.value.errors == [ErrorItem(message=err, instance_path='')]


def test_decimal_validation__invalid_type():
    s = Serializer(Decimal)
    with pytest.raises(SchemaValidationError) as e:
        s.load('foo')

    assert e.value.errors == [ErrorItem(message='"foo" is not of type "decimal"', instance_path='')]


def test_boolean_validation__invalid_type():
    s = Serializer(bool)
    _check_errors(s, 'foo', [ErrorItem(message='"foo" is not of type "boolean"', instance_path='')])


def test_uuid_validation__invalid_type():
    s = Serializer(uuid.UUID)
    with pytest.raises(SchemaValidationError) as e:
        s.load('foo')

    assert e.value.errors == [ErrorItem(message='"foo" is not of type "uuid"', instance_path='')]


def test_time_validation__invalid_type():
    s = Serializer(time)
    with pytest.raises(SchemaValidationError) as e:
        s.load('foo')

    assert e.value.errors == [ErrorItem(message='"foo" is not of type "time"', instance_path='')]


def test_datetime_validation__invalid_type():
    s = Serializer(datetime)
    with pytest.raises(SchemaValidationError) as e:
        s.load('foo')

    assert e.value.errors == [ErrorItem(message='"foo" is not of type "datetime"', instance_path='')]


def test_date_validation__invalid_type():
    s = Serializer(date)
    with pytest.raises(SchemaValidationError) as e:
        s.load('foo')

    assert e.value.errors == [ErrorItem(message='"foo" is not of type "date"', instance_path='')]


def test_dataclass_validation__invalid_type():
    @dataclass
    class A:
        a: int

    s = Serializer(A)
    with pytest.raises(SchemaValidationError) as e:
        s.load('foo')

    assert e.value.errors == [ErrorItem(message='"foo" is not of type "object"', instance_path='')]


def test_dataclass_validation__missing_field():
    @dataclass
    class A:
        a: int

    s = Serializer(A)
    _check_errors(s, {}, [ErrorItem(message='"a" is a required property', instance_path='a')])


def test_dataclass_validation__missing_field__with_instance_path():
    @dataclass
    class Foo:
        a: int

    @dataclass
    class Bar:
        foo: Foo

    s = Serializer(Bar)
    _check_errors(s, {'foo': {}}, [ErrorItem(message='"a" is a required property', instance_path='foo/a')])


def test_typed_dict_validation__invalid_type():
    class A(TypedDict):
        a: int

    s = Serializer(A)
    with pytest.raises(SchemaValidationError) as e:
        s.load('foo')

    assert e.value.errors == [ErrorItem(message='"foo" is not of type "dict"', instance_path='')]


def test_typed_dict_validation__missing_field():
    class A(TypedDict):
        a: int

    s = Serializer(A)
    _check_errors(s, {}, [ErrorItem(message='"a" is a required property', instance_path='a')])


def test_typed_dict_validation__missing_field__with_instance_path():
    class Foo(TypedDict):
        a: int

    class Bar(TypedDict):
        foo: Foo

    s = Serializer(Bar)
    _check_errors(s, {'foo': {}}, [ErrorItem(message='"a" is a required property', instance_path='foo/a')])


def test_list_validation__invalid_type():
    s = Serializer(list[int])
    _check_errors(s, 'foo', [ErrorItem(message='"foo" is not of type "list"', instance_path='')])


def test_list_validation__invalid_item_type():
    s = Serializer(list[int])
    _check_errors(s, [2, 3, 'foo'], [ErrorItem(message='"foo" is not of type "integer"', instance_path='2')])


def test_enum_validation__invalid_type():
    class A(Enum):
        A = 1
        B = 'foo'

    s = Serializer(A)
    _check_errors(s, 'bar', [ErrorItem(message='"bar" is not one of [1, "foo"]', instance_path='')])


def test_dict_validation__invalid_type():
    s = Serializer(dict[str, int])
    _check_errors(s, 'foo', [ErrorItem(message='"foo" is not of type "dict"', instance_path='')])


def test_dict_validation__invalid_value_type():
    s = Serializer(dict[str, int])
    _check_errors(s, {'foo': 1, 'bar': '2'}, [ErrorItem(message='"2" is not of type "integer"', instance_path='bar')])


def test_dict_validation__invalid_key_type():
    s = Serializer(dict[str, int])
    with pytest.raises(SchemaValidationError) as e:
        s.load({1: 1})

    assert e.value.errors == [ErrorItem(message='1 is not of type "string"', instance_path='1')]


def test_tuple_validation__invalid_type():
    s = Serializer(tuple[int, str])
    with pytest.raises(SchemaValidationError) as e:
        s.load('foo')

    assert e.value.errors == [ErrorItem(message='"foo" is not of type "sequence"', instance_path='')]


def test_tuple_validation__invalid_item_type():
    s = Serializer(tuple[int, str])
    _check_errors(s, [1, 2], [ErrorItem(message='2 is not of type "string"', instance_path='1')])


def test_tuple_validation__invalid_length():
    s = Serializer(tuple[int, str])

    with pytest.raises(SchemaValidationError) as e:
        s.load([1])

    assert e.value.errors == [ErrorItem(message='[1] has less than 2 items', instance_path='')]

    with pytest.raises(SchemaValidationError) as e:
        s.load([1, 'foo', 3])

    assert e.value.errors == [ErrorItem(message="[1, 'foo', 3] has more than 2 items", instance_path='')]


def test_bytes_validation__invalid_type():
    s = Serializer(bytes)
    with pytest.raises(SchemaValidationError) as e:
        s.load('foo')

    assert e.value.errors == [ErrorItem(message='"foo" is not of type "bytes"', instance_path='')]


@dataclass
class A:
    type: Literal['A']
    a: int


@dataclass
class B:
    type: Literal['B']
    b: str


def test_tagged_union_validation__invalid_type():
    s = Serializer(Annotated[Union[A, B], Discriminator('type')])
    with pytest.raises(SchemaValidationError) as e:
        s.load('foo')

    assert e.value.errors == [ErrorItem(message='"foo" is not of type "dict"', instance_path='')]


def test_tagged_union_validation__invalid_discriminator():
    s = Serializer(Annotated[Union[A, B], Discriminator('type')])
    with pytest.raises(SchemaValidationError) as e:
        s.load({'type': 'C'})

    assert e.value.errors == [
        ErrorItem(message='"C" is not one of ["A", "B"] discriminator values', instance_path='type')
    ]


def test_tagged_union_validation__discriminator_missing():
    s = Serializer(Annotated[Union[A, B], Discriminator('type')])
    with pytest.raises(SchemaValidationError) as e:
        s.load({})

    assert e.value.errors == [ErrorItem(message='"type" is a required property', instance_path='type')]


def test_dump_tagged_union_validation__discriminator_missing():
    s = Serializer(Annotated[Union[A, B], Discriminator('type')])
    with pytest.raises(SchemaValidationError) as e:
        s.dump({})

    assert e.value.errors == [ErrorItem(message='"type" is a required property', instance_path='type')]


def test_literal_validation__invalid_value():
    s = Serializer(Literal['foo', 'bar'])
    with pytest.raises(SchemaValidationError) as e:
        s.load(1)

    assert e.value.errors == [ErrorItem(message='1 is not one of ["bar", "foo"]', instance_path='')]


def test_instance_path():
    @dataclass
    class Foo:
        a: int

    @dataclass
    class Bar:
        foo: tuple[Foo]

    @dataclass
    class Baz:
        bar: list[Bar]

    s = Serializer(Baz)
    _check_errors(
        s,
        {'bar': [{'foo': [{'a': 1}]}, {'foo': [{'b': 1}]}]},
        [ErrorItem(message='"a" is a required property', instance_path='bar/1/foo/0/a')],
    )



"""Query params in most cases, field values are of string type"""

from dataclasses import dataclass
from typing import Annotated, Literal, TypedDict, Union
from enum import Enum

import pytest
from multidict import MultiDict

import serpyco_rs
from serpyco_rs import metadata


def test_load_query_params__flat_type__exception():
    serializer = serpyco_rs.Serializer(int)
    with pytest.raises(RuntimeError) as exc_info:
        serializer.load_query_params(MultiDict({'id': 'a'}))

    assert str(exc_info.value) == 'This type is not deserializable from query params'


def test_load_query_params__int():
    @dataclass
    class Foo:
        id: int

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict({'id': '1'})) == Foo(id=1)


def test_load_query_params__int__empty():
    @dataclass
    class Foo:
        id: int

    serializer = serpyco_rs.Serializer(Foo)
    with pytest.raises(serpyco_rs.SchemaValidationError):
        serializer.load_query_params(MultiDict())


def test_load_query_params__int__invalid():
    @dataclass
    class Foo:
        id: int

    serializer = serpyco_rs.Serializer(Foo)
    with pytest.raises(serpyco_rs.SchemaValidationError):
        serializer.load_query_params(MultiDict({'id': '1.1'}))


def test_load_query_params__float():
    @dataclass
    class Foo:
        id: float

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict({'id': '1.1'})) == Foo(id=1.1)


def test_load_query_params__float__invalid():
    @dataclass
    class Foo:
        id: float

    serializer = serpyco_rs.Serializer(Foo)
    with pytest.raises(serpyco_rs.SchemaValidationError):
        serializer.load_query_params(MultiDict({'id': '1.1.1'}))


def test_load_query_params__bool():
    @dataclass
    class Foo:
        f: bool

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict({'f': 't'})) == Foo(f=True)
    assert serializer.load_query_params(MultiDict({'f': 'T'})) == Foo(f=True)
    assert serializer.load_query_params(MultiDict({'f': 'true'})) == Foo(f=True)
    assert serializer.load_query_params(MultiDict({'f': 'True'})) == Foo(f=True)
    assert serializer.load_query_params(MultiDict({'f': 'TRUE'})) == Foo(f=True)

    assert serializer.load_query_params(MultiDict({'f': 'f'})) == Foo(f=False)
    assert serializer.load_query_params(MultiDict({'f': 'F'})) == Foo(f=False)
    assert serializer.load_query_params(MultiDict({'f': 'false'})) == Foo(f=False)
    assert serializer.load_query_params(MultiDict({'f': 'False'})) == Foo(f=False)
    assert serializer.load_query_params(MultiDict({'f': 'FALSE'})) == Foo(f=False)


def test_load_query_params__bool__invalid():
    @dataclass
    class Foo:
        f: bool

    serializer = serpyco_rs.Serializer(Foo)
    with pytest.raises(serpyco_rs.SchemaValidationError):
        serializer.load_query_params(MultiDict({'f': '1'}))


def test_load_query_params__list():
    @dataclass
    class Foo:
        ids: list[int]

    serializer = serpyco_rs.Serializer(Foo)
    data = MultiDict([('ids', '1'), ('ids', 2), ('ids', '3')])
    assert serializer.load_query_params(data) == Foo(ids=[1, 2, 3])


def test_load_query_params__list__invalid():
    @dataclass
    class Foo:
        ids: list[int]

    serializer = serpyco_rs.Serializer(Foo)
    with pytest.raises(serpyco_rs.SchemaValidationError):
        serializer.load_query_params(MultiDict([('ids', 1), ('ids', '2.12'), ('ids', 3)]))


def test_load_query_params__list__empty():
    @dataclass
    class Foo:
        ids: list[int]

    serializer = serpyco_rs.Serializer(Foo)
    with pytest.raises(serpyco_rs.SchemaValidationError):
        serializer.load_query_params(MultiDict())


def test_load_query_params__list__optional():
    @dataclass
    class Foo:
        ids: Union[list[int], None] = None

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict()) == Foo(ids=None)


def test_load_query_params__tuple():
    @dataclass
    class Foo:
        ids: tuple[int, bool]

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict([('ids', '1'), ('ids', 'true')])) == Foo(ids=(1, True))


def test_load_query_params__tuple__invalid():
    @dataclass
    class Foo:
        ids: tuple[int, bool]

    serializer = serpyco_rs.Serializer(Foo)
    with pytest.raises(serpyco_rs.SchemaValidationError):
        serializer.load_query_params(MultiDict([('ids', '1')]))


def test_load_query_params__union():
    @dataclass
    class Foo:
        f: Union[int, bool]

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict({'f': '1'})) == Foo(f=1)
    assert serializer.load_query_params(MultiDict({'f': 'true'})) == Foo(f=True)


def test_load_query_params__custom_encoder():
    @dataclass
    class Foo:
        id: Annotated[int, metadata.deserialize_with(lambda value: int(value.split('-')[1]))]

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict({'id': '1-2'})) == Foo(id=2)


def test_load_query_params__custom_encoder_list():
    @dataclass
    class Foo:
        id: Annotated[list[int], metadata.deserialize_with(lambda value: [int(v.split('-')[1]) for v in value])]

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict([('id', '1-2'), ('id', '3-4')])) == Foo(id=[2, 4])


class FooEnum(Enum):
    A = 'a'
    B = 2


def test_load_query_params__enum():

    @dataclass
    class Foo:
        e: FooEnum

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict({'e': 'a'})) == Foo(e=FooEnum.A)
    assert serializer.load_query_params(MultiDict({'e': '2'})) == Foo(e=FooEnum.B)


def test_load_query_params__literal():

    @dataclass
    class Foo:
        e: Literal['a', 2]

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict({'e': 'a'})) == Foo(e='a')
    assert serializer.load_query_params(MultiDict({'e': '2'})) == Foo(e=2)


def test_load_query_params__typed_dict():

    class Foo(TypedDict):
        id: int
        name: str

    serializer = serpyco_rs.Serializer(Foo)
    assert serializer.load_query_params(MultiDict({'id': '1', 'name': 'test'})) == Foo(id=1, name='test')


def test_load_query_params__dict():
    serializer = serpyco_rs.Serializer(dict[str, int])
    assert serializer.load_query_params(MultiDict({'id': '1'})) == {'id': 1}


def test_load_query_params__dict_of_lists():
    serializer = serpyco_rs.Serializer(dict[str, list[int]])
    assert serializer.load_query_params(MultiDict([('foo', '1'), ('foo', '2'), ('bar', '3')])) == {
        'foo': [1, 2],
        'bar': [3],
    }



import sys
from functools import singledispatch
from typing import Any, Optional, Union, cast

if sys.version_info >= (3, 10):
    from typing import TypeGuard
else:
    from typing_extensions import TypeGuard

from .. import _describe as describe
from ._entities import (
    ArrayType,
    Boolean,
    DiscriminatedUnionType,
    Discriminator,
    IntegerType,
    Null,
    NumberType,
    ObjectType,
    RefType,
    Schema,
    StringType,
)


def get_json_schema(t: describe.BaseType) -> dict[str, Any]:
    schema = to_json_schema(t)
    definitions: dict[str, Any] = {}
    schema_def = schema.dump(definitions)
    components = {'components': {'schemas': definitions}} if definitions else {}
    return {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        **schema_def,
        **components,
    }


@singledispatch
def to_json_schema(_: Any, doc: Optional[str] = None) -> Schema:
    return Schema(description=doc)


@to_json_schema.register
def _(arg: describe.StringType, doc: Optional[str] = None) -> Schema:
    return StringType(
        minLength=arg.min_length,
        maxLength=arg.max_length,
        description=doc,
    )


@to_json_schema.register
def _(arg: describe.IntegerType, doc: Optional[str] = None) -> Schema:
    return IntegerType(
        minimum=arg.min,
        maximum=arg.max,
        description=doc,
    )


@to_json_schema.register
def _(_: describe.BytesType, doc: Optional[str] = None) -> Schema:
    return StringType(
        format='binary',
        description=doc,
    )


@to_json_schema.register
def _(arg: describe.FloatType, doc: Optional[str] = None) -> Schema:
    return NumberType(
        minimum=arg.min,
        maximum=arg.max,
        description=doc,
    )


@to_json_schema.register
def _(_: describe.DecimalType, doc: Optional[str] = None) -> Schema:
    return Schema(
        oneOf=[
            StringType(
                format='decimal',
            ),
            NumberType(
                format='decimal',
            ),
        ],
        description=doc,
    )


@to_json_schema.register
def _(_: describe.BooleanType, doc: Optional[str] = None) -> Schema:
    return Boolean()


@to_json_schema.register
def _(_: describe.UUIDType, doc: Optional[str] = None) -> Schema:
    return StringType(
        format='uuid',
        description=doc,
    )


@to_json_schema.register
def _(_: describe.TimeType, doc: Optional[str] = None) -> Schema:
    return StringType(
        format='time',
        description=doc,
    )


@to_json_schema.register
def _(_: describe.DateTimeType, doc: Optional[str] = None) -> Schema:
    return StringType(
        format='date-time',
        description=doc,
    )


@to_json_schema.register
def _(_: describe.DateType, doc: Optional[str] = None) -> Schema:
    return StringType(
        format='date',
        description=doc,
    )


@to_json_schema.register
def _(arg: describe.EnumType, doc: Optional[str] = None) -> Schema:
    return Schema(
        enum=[item.value for item in arg.cls],
        description=doc,
    )


@to_json_schema.register
def _(arg: describe.OptionalType, doc: Optional[str] = None) -> Schema:
    return Schema(
        anyOf=[
            Null(),
            to_json_schema(arg.inner),
        ],
        description=doc,
    )


@to_json_schema.register
def _(arg: describe.EntityType, doc: Optional[str] = None) -> Schema:
    return ObjectType(
        properties={prop.dict_key: to_json_schema(prop.field_type, prop.doc) for prop in arg.fields},
        required=[prop.dict_key for prop in arg.fields if prop.required] or None,
        name=arg.name,
        description=arg.doc,
    )


@to_json_schema.register
def _(arg: describe.TypedDictType, doc: Optional[str] = None) -> Schema:
    return ObjectType(
        properties={prop.dict_key: to_json_schema(prop.field_type, prop.doc) for prop in arg.fields},
        required=[prop.dict_key for prop in arg.fields if prop.required] or None,
        name=arg.name,
        description=arg.doc,
    )


@to_json_schema.register
def _(arg: describe.ArrayType, doc: Optional[str] = None) -> Schema:
    return ArrayType(
        items=to_json_schema(arg.item_type),
        description=doc,
    )


@to_json_schema.register
def _(arg: describe.DictionaryType, doc: Optional[str] = None) -> Schema:
    return ObjectType(
        additionalProperties=to_json_schema(arg.value_type),
        description=doc,
    )


@to_json_schema.register
def _(arg: describe.TupleType, doc: Optional[str] = None) -> Schema:
    return ArrayType(
        prefixItems=[to_json_schema(item) for item in arg.item_types],
        minItems=len(arg.item_types),
        maxItems=len(arg.item_types),
        description=doc,
    )


@to_json_schema.register
def _(_: describe.AnyType, doc: Optional[str] = None) -> Schema:
    return Schema(description=doc)


@to_json_schema.register
def _(holder: describe.RecursionHolder, doc: Optional[str] = None) -> Schema:
    return RefType(description=doc, ref=f'#/components/schemas/{holder.name}')


@to_json_schema.register
def _(arg: describe.LiteralType, doc: Optional[str] = None) -> Schema:
    return Schema(
        enum=list(arg.args),
        description=doc,
    )


@to_json_schema.register
def _(arg: describe.UnionType, doc: Optional[str] = None) -> Schema:
    return Schema(
        oneOf=[to_json_schema(t) for t in arg.item_types],
        description=doc,
    )


@to_json_schema.register
def _(arg: describe.DiscriminatedUnionType, doc: Optional[str] = None) -> Schema:
    objects = {
        name: schema
        for name, t in arg.item_types.items()
        if (schema := to_json_schema(t)) and _check_unions_schema_types(schema)
    }

    return DiscriminatedUnionType(
        oneOf=list(objects.values()),
        discriminator=Discriminator(
            property_name=arg.load_discriminator,
            mapping={name: cast(str, val.ref) for name, val in objects.items()},
        ),
        description=doc,
    )


def _check_unions_schema_types(schema: Schema) -> TypeGuard[Union[ObjectType, RefType]]:
    if isinstance(schema, (ObjectType, RefType)):
        return True
    raise RuntimeError(f'Unions schema items must be ObjectType or RefType. Current: {schema}')



from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any, Union
from typing import Optional, Annotated

from serpyco_rs.metadata import Alias
import serpyco_rs

from ._utils import get_dataclass_args


class IssueState(Enum):
    OPEN = "open"
    CLOSED = "closed"


class MilestoneState(Enum):
    OPEN = "open"
    CLOSED = "closed"


class IssueStateReason(Enum):
    COMPLETED = "completed"
    REOPENED = "reopened"
    NOT_PLANNED = "not_planned"


class AuthorAssociation(Enum):
    COLLABORATOR = "COLLABORATOR"
    CONTRIBUTOR = "CONTRIBUTOR"
    FIRST_TIMER = "FIRST_TIMER"
    FIRST_TIME_CONTRIBUTOR = "FIRST_TIME_CONTRIBUTOR"
    MANNEQUIN = "MANNEQUIN"
    MEMBER = "MEMBER"
    NONE = "NONE"
    OWNER = "OWNER"


@dataclass(**get_dataclass_args())
class User:
    login: str
    id: int
    node_id: str
    avatar_url: str
    gravatar_id: Optional[str]
    url: str
    html_url: str
    followers_url: str
    following_url: str
    gists_url: str
    starred_url: str
    subscriptions_url: str
    organizations_url: str
    repos_url: str
    events_url: str
    received_events_url: str
    type: str
    site_admin: bool
    name: Optional[str] = None
    email: Optional[str] = None
    starred_at: Optional[datetime] = None


@dataclass(**get_dataclass_args())
class IssueLabel:
    id: int
    node_id: str
    url: str
    name: str
    description: Optional[str]
    color: Optional[str]
    default: bool


@dataclass(**get_dataclass_args())
class Milestone:
    url: str
    html_url: str
    labels_url: str
    id: int
    node_id: str
    number: int
    title: str
    description: Optional[str]
    creator: Optional[User]
    open_issues: int
    closed_issues: int
    created_at: datetime
    updated_at: datetime
    closed_at: Optional[datetime]
    due_on: Optional[datetime]
    state: MilestoneState = MilestoneState.OPEN


@dataclass(**get_dataclass_args())
class Reactions:
    url: str
    total_count: int
    plus_one: Annotated[int, Alias('+1')]
    minus_one: Annotated[int, Alias('-1')]
    laugh: int
    confused: int
    heart: int
    hooray: int
    eyes: int
    rocket: int


@dataclass(**get_dataclass_args())
class Issue:
    id: int
    node_id: str
    url: str
    repository_url: str
    labels_url: str
    comments_url: str
    events_url: str
    html_url: str
    number: int
    state: IssueState
    state_reason: Optional[IssueStateReason]
    title: str
    body: Optional[str]
    user: Optional[User]
    labels: list[Union[IssueLabel, str]]
    assignee: Optional[User]
    assignees: Optional[list[User]]
    milestone: Optional[Milestone]
    locked: bool
    active_lock_reason: Optional[str]
    comments: int
    closed_at: Optional[datetime]
    created_at: datetime
    updated_at: datetime
    closed_by: Optional[User]
    author_association: AuthorAssociation
    draft: bool = False
    body_html: Optional[str] = None
    body_text: Optional[str] = None
    timeline_url: Optional[str] = None
    reactions: Optional[Reactions] = None


_serializer = serpyco_rs.Serializer(Issue)


def load(data: dict[str, Any]) -> Issue:
    return _serializer.load(data)


def dump(obj: Issue) -> dict[str, Any]:
    return _serializer.dump(obj)



import gc
import sys
from dataclasses import dataclass
from decimal import Decimal
from enum import Enum
from typing import Any

import serpyco_rs


def test_repr_ref_count_segfault_on_default_value():
    class Bar(Enum):
        a = 'a'

    @dataclass
    class Foo:
        lang: Bar = Bar.a

    serializer = serpyco_rs.Serializer(Foo)

    for _i in range(10):
        serializer.load({})
        gc.collect()


def test_load_rc_count_decimal():
    serializer = serpyco_rs.Serializer(Decimal)

    val = '123.1'

    ref_count = sys.getrefcount(val)

    for _i in range(100):
        serializer.load(val)
        gc.collect()

    assert ref_count == sys.getrefcount(val)


def test_load_rc_count_dict_of_decimals():
    serializer = serpyco_rs.Serializer(dict[str, Decimal])
    val = {'a_foo': 5000, 'b_foo': 5000.0, 'c_foo': '60000'}
    ref_counts = _get_dict_rc_count(val)

    for _i in range(100):
        serializer.load(val)
        gc.collect()

    assert ref_counts == _get_dict_rc_count(val)


def test_dump_rc_count_decimal():
    serializer = serpyco_rs.Serializer(Decimal)

    val = Decimal('123.1')

    ref_count = sys.getrefcount(val)

    for _i in range(100):
        serializer.dump(val)
        gc.collect()

    assert ref_count == sys.getrefcount(val)


def _get_dict_rc_count(val: dict[Any, Any]) -> dict[str, int]:
    return {
        'keys': [sys.getrefcount(k) for k in val],
        'vals': [sys.getrefcount(v) for v in val.values()],
    }



from dataclasses import dataclass
from typing import Any, Optional

from serpyco_rs._impl import BaseType
from serpyco_rs.metadata import FieldFormat, NoneAsDefaultForOptional, NoneFormat


@dataclass(frozen=True, unsafe_hash=True)
class MetaStateKey:
    cls: type
    field_format: FieldFormat
    none_format: NoneFormat
    none_as_default_for_optional: NoneAsDefaultForOptional


@dataclass
class Meta:
    globals: dict[str, Any]
    state: dict[MetaStateKey, Optional[BaseType]]
    discriminator_field: Optional[str] = None

    def add_to_state(self, key: MetaStateKey, value: Optional[BaseType]) -> None:
        self.state[key] = value

    def __getitem__(self, item: MetaStateKey) -> Optional[BaseType]:
        return self.state.get(item)

    def has_in_state(self, key: MetaStateKey) -> bool:
        return key in self.state



import abc
from typing import Annotated, Any, Generic, Protocol, TypeVar, Union, cast, overload

from ._describe import describe_type
from ._impl import Serializer as _Serializer
from ._json_schema import get_json_schema
from .metadata import CamelCase, ForceDefaultForOptional, OmitNone

_T = TypeVar('_T', bound=Any)
_D = TypeVar('_D')


class _MultiMapping(Protocol[_T, _D]):
    """Protocol for a multi-mapping type."""

    @abc.abstractmethod
    def __getitem__(self, __key: str) -> _T: ...

    @overload
    @abc.abstractmethod
    def getall(self, key: str) -> list[_T]: ...
    @overload
    @abc.abstractmethod
    def getall(self, key: str, default: _D) -> Union[list[_T], _D]: ...
    @abc.abstractmethod
    def getall(self, key: str, default: _D = ...) -> Union[list[_T], _D]: ...


class Serializer(Generic[_T]):
    def __init__(
        self,
        t: type[_T],
        *,
        camelcase_fields: bool = False,
        omit_none: bool = False,
        force_default_for_optional: bool = False,
        naive_datetime_to_utc: bool = False,
    ) -> None:
        """
        Create a serializer for the given type.

        :param t: The type to serialize/deserialize.
        :param camelcase_fields: If True, the serializer will convert field names to camelCase.
        :param omit_none: If True, the serializer will omit None values from the output.
        :param force_default_for_optional: If True, the serializer will force default values for optional fields.
        :param naive_datetime_to_utc: If True, the serializer will convert naive datetimes to UTC.
        """
        if camelcase_fields:
            t = cast(type[_T], Annotated[t, CamelCase])
        if omit_none:
            t = cast(type(_T), Annotated[t, OmitNone])  # type: ignore
        if force_default_for_optional:
            t = cast(type(_T), Annotated[t, ForceDefaultForOptional])  # type: ignore
        type_info = describe_type(t)
        self._schema = get_json_schema(type_info)
        self._encoder: _Serializer[_T] = _Serializer(type_info, naive_datetime_to_utc)

    def dump(self, value: _T) -> Any:
        """Serialize the given value to a JSON-serializable object.

        :param value: The value to serialize.
        """
        return self._encoder.dump(value)

    def load(self, data: Any) -> _T:
        """Deserialize the given JSON-serializable object to the target type.

        :param data: The data to deserialize.
        """
        return self._encoder.load(data)

    def load_query_params(self, data: _MultiMapping[Any, Any]) -> _T:
        """Deserialize the given query parameters to the target type.

        :param data: The query parameters to deserialize.
        """
        return self._encoder.load_query_params(data)

    def get_json_schema(self) -> dict[str, Any]:
        """Get the JSON schema for the target type."""
        return self._schema
